import cmd  # https://docs.python.org/3/library/cmd.html
from typing import IO, Optional

from src.ergo_cli import ErgoCli


class ErgoCmd(cmd.Cmd):
    intro: str = ''
    prompt: str = ''

    def __init__(self, cli: ErgoCli, completekey: str = 'tab', stdin: Optional[IO[str]] = None, stdout: Optional[IO[str]] = None) -> None:
        super().__init__(completekey, stdin, stdout)
        self._cli = cli
        ErgoCmd.intro = self._cli.intro
        ErgoCmd.prompt = self._cli.prompt

    # def onecmd(self, line):
    #   try:
    #     return super().onecmd(line)
    #   except Exception as err:
    #     print(f'*** {err}')
    #     traceback.print_exc(file=sys.stdout)
    #     return False  # don't stop

    def do_exit(self, line: str) -> bool:
        return True

    def do_run(self, line: str) -> bool:
        args = line.split()
        return self._cli.run(args[0], *args[1:])

    def do_http(self, line: str) -> bool:
        args = line.split()
        return self._cli.http(args[0], *args[1:])



from flask import Flask, request  # , abort

from src.http_invoker import HttpInvoker
from src.payload import Payload


class FlaskHttpInvoker(HttpInvoker):
    def start(self) -> int:
        app = Flask(__name__)

        @app.route(self.route)
        def handler() -> str:
            data_out: Payload = Payload()
            data_in: Payload = Payload(dict(request.args))
            # data_in(f'route: {str(request.url_rule)}')
            # try:
            self._invocable.invoke(data_out, data_in)
            # except Exception as err:
            #     print(err)
            #     abort(400)

            return str(data_out)

        app.run(host='0.0.0.0', port=self._port)
        return False


# from werkzeug.serving import make_server
# import threading
# from multiprocessing import Process
# class ServerThread(threading.Thread):

#     def __init__(self, app):
#         threading.Thread.__init__(self)
#         self.srv = make_server('127.0.0.1', 5000, app)
#         self.ctx = app.app_context()
#         self.ctx.push()

#     def run(self):
#         print('starting server')
#         self.srv.serve_forever()

#     def shutdown(self):
#         self.srv.shutdown()

# class FlaskHttpInvoker(HttpInvoker):
#   def start(self) -> int:
#     global server
#     app = Flask('myapp')
#     #...
#     server = ServerThread(app)
#     server.start()
#     print('server started')

#   def stop(self):
#       global server
#       server = Process(target='myapp')
#       server.terminate()
#       server.join()



import datetime
import os

from colors import color

from src.flask_http_invoker import FlaskHttpInvoker
from src.function_invocable import FunctionInvocable
from src.payload import Payload
from src.version import get_version


def format_date(sec: float) -> str:
    dtf: str = '%b %d %Y, %H:%M:%S.%f'
    return datetime.datetime.fromtimestamp(sec).strftime(dtf)[:-3]


def get_version_path() -> str:
    return os.path.dirname(os.path.abspath(__file__)) + '/version.py'


class ErgoCli:
    @property
    def prompt(self) -> str:
        return f'{color("ergo", fg="#33ff33")} {color("∴", fg="#33ff33")} '

    @property
    def intro(self) -> str:
        version: str = get_version()
        timestamp: str = format_date(os.path.getmtime(get_version_path()))
        intro: str = f'ergo {version} ({timestamp})\nType help or ? to list commands.'

        return str(color(intro, fg='#ffffff'))

    def run(self, ref: str, *args: str) -> bool:
        try:
            result: Payload = Payload()
            host = FunctionInvocable(ref)
            host.invoke(result, Payload(dict(zip([str(i) for i in range(len(args))], args))))
            print(str(result))
        except Exception as err:
            print(f'*** {err}')
            raise err
        return False

    def http(self, ref: str, *args: str) -> bool:
        host = FlaskHttpInvoker(FunctionInvocable(ref))
        host.start()
        return False



from typing import Tuple

import click
from click_default_group import \
    DefaultGroup  # https://pypi.org/project/click-default-group/

from src.ergo_cli import ErgoCli
from src.ergo_cmd import ErgoCmd

ERGO_CLI = ErgoCli()


@click.group(cls=DefaultGroup, default='shell', default_if_no_args=True)
def main() -> bool:
    pass


@main.command()
def shell() -> bool:
    ErgoCmd(ERGO_CLI).cmdloop()
    return False


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def run(ref: str, arg: Tuple[str]) -> bool:
    return ERGO_CLI.run(ref, *list(arg))


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def http(ref: str, arg: Tuple[str]) -> bool:
    return ERGO_CLI.http(ref, *list(arg))



from typing import Dict, List, Optional


class Payload:
    def __init__(self, data: Optional[Dict[str, str]] = None) -> None:
        self._data: Dict[str, str] = data or {}

    def get(self, key: str) -> Optional[str]:
        return self._data.get(key)

    def set(self, key: str, value: str) -> None:
        self._data[key] = value

    def list(self) -> List[str]:
        return list(self._data.values())

    def __str__(self):
      return str(self._data)